#!/usr/bin/env python3
from mdbook_pdf_summary.parse_section import Section, parse_section_tree
import pypdf
import lxml.html
import urllib
import argparse
import os


def get_dom_id(node: Section):
    """
    Get the DOM ID in `print.html`.

    ATTENTION:
    Before calling `get_href_path`, the user must use `check_title.check_title` to make sure the title is matched.

    The rule is:
    1. DOM="${MARKDOWN PATH}"
    2. lowercase all the characters.
    3. replace all the `/` with `-`.
    4. replace all the ` ` with `-`

    e.g.
    line in `SUMMARY.md`(double G not typoo):
    - [Getting Started](./user_guide/GGetting_started.md)

    line in `print.html`:
    href="#user_guide-ggetting_started-getting-started"
    """
    source_path = node.source_file
    # remove the prefix `./`
    if source_path.startswith("./"):
        source_path = source_path[2:]
    # strip the suffix
    source_path = source_path.split(".")[0]
    result = source_path
    # lowercase all the characters
    result = result.lower()
    # replace the `/` with `-`
    result = result.replace("/", "-")
    # replace the ` ` with `-`
    result = result.replace(" ", "-")
    return result


def add_outline(
    html_root, reader: pypdf.PdfReader, writer: pypdf.PdfWriter, node: Section
):
    """
    Add outline to the PDF file.

    Parameters:
    -----------
    html_root: lxml.html.HtmlElement
        The root element of the HTML file.
    reader: pypdf.PdfReader
        The reader of the PDF file.
    writer: pypdf.PdfWriter
        The writer of the PDF file.
    node: Section
        The node of the section tree.
    """
    if not node.is_root():
        id = get_dom_id(node)
        try:
            results = html_root.get_element_by_id(id)
        except KeyError:
            print("[ERROR] Element not found: [{}]".format(id))
            return

        if results is None:
            print("[ERROR] Element is None, id: [{}]".format(id))
            return
        dest = reader.named_destinations["/{}".format(urllib.parse.quote(id))]

        page = None
        fit = None
        if dest.get("/Type") != "/Fit":
            page = reader.get_destination_page_number(dest)
            fit = pypdf.generic.Fit(
                dest.get("/Type"),
                (dest.get("/Left"), dest.get("/Top"), dest.get("/Zoom")),
            )
        node.outline_item = writer.add_outline_item(
            str(node), page, node.parent.outline_item, fit=fit
        )
    # dfs
    for child in node.children:
        add_outline(html_root, reader, writer, child)


def main():
    parser = argparse.ArgumentParser(
        prog="mdbook_pdf_summary", description="Add outline to the PDF file."
    )
    parser.add_argument(
        "--html_path",
        type=str,
        help="path of the `print.html` generated `mdbook-pdf`",
        default="book/html/print.html",
    )
    parser.add_argument(
        "--pdf_path",
        type=str,
        help="path of the `output.pdf` generated `mdbook-pdf`",
        default="book/pdf/output.pdf",
    )
    parser.add_argument(
        "--summary_path",
        type=str,
        help="path of the `SUMMARY.md`",
        default="src/SUMMARY.md",
    )
    parser.add_argument(
        "--output_path",
        type=str,
        help="path of the output PDF file",
        default="output_with_outline.pdf",
    )
    args = parser.parse_args()
    print("============ args =============")
    print("args.html_path: ", args.html_path)
    print("args.pdf_path: ", args.pdf_path)
    print("args.summary_path: ", args.summary_path)
    print("args.output_path: ", args.output_path)
    if not os.path.exists(args.html_path):
        raise FileNotFoundError(f"{args.html_path} does not exist")
    if not os.path.exists(args.pdf_path):
        raise FileNotFoundError(f"{args.pdf_path} does not exist")
    if not os.path.exists(args.summary_path):
        raise FileNotFoundError(f"{args.summary_path} does not exist")

    reader = pypdf.PdfReader(args.pdf_path)
    writer = pypdf.PdfWriter()
    writer.append(reader)
    with open(args.summary_path) as f:
        md_text = f.read()
    section_root = parse_section_tree(md_text)
    html_root = None
    with open(args.html_path, "r", encoding="utf8") as f:
        data = f.read()
        html_root = lxml.html.fromstring(data)
    if html_root is None:
        raise ("[ERROR] html_root is None")
    add_outline(html_root, reader, writer, section_root)
    with open(args.output_path, "wb") as f:
        writer.write(f)
        print("[INFO] Write to {}".format(args.output_path))



from mdbook_pdf_summary import parse_section_tree, print_section_tree
if __name__ == "__main__":
    summary_path = "./example/a_book/src/SUMMARY.md"
    summary_path = "./Interface/src/SUMMARY.md"
    with open(summary_path) as f:
        md_text = f.read()
    result = parse_section_tree(md_text)
    print_section_tree(result)


import re    
class Section:
    def __init__(self, title: str, source_file: str, depth: int, index: int):
        self.title = title
        self.source_file = source_file
        self.depth = depth
        self.index = index
        self.parent = None
        self.children = []
        self.outline_item = None
    def set_parent(self, parent):
        self.parent = parent
    def add_children(self, child):
        self.children.append(child)
    def path_to_root(self):
        path = []
        node = self
        while not node.is_root():
            path.append(str(node.index + 1))
            node = node.parent
        path = path[::-1]
        return path
    def is_root(self):
        return self.parent is None

    def __str__(self):
        path = self.path_to_root()
        return "{}. {}".format(".".join(path), self.title)

def print_section_tree(root: Section):
    print(root)
    for child in root.children:
        print_section_tree(child)

def parse_section_tree(md_text: str):
    """
    Construct a tree from markdown unordered list.
    
    Markdown format:
    - Chaptor 1
        - Section 1.1
        - Section 1.2
            - Section 1.2.1
    - Chaptor 2
    - Chaptor 3
        - Section 3.1

    Consturct a tree need the line order and the indent number.
    """

    # root.depth = 0
    root = Section("root", "", 0, 0)
    # key: depth, value: node.
    bfs_map = {0: [root]}
    # can just match unordered_list, a.k.a. `-`
    pattern = re.compile(r'( *)- ([^:\n]+)(?:: ([^\n]*))?\n?')
    tmp = None
    min_indent_num = 4
    for indent, name, value in pattern.findall(md_text):
        # `-[${section_name}](${md_path})``
        title = name.split("](")[0].split("[")[1]
        source_file = name.split("](")[1].split(")")[0]
        indent_num = len(indent)
        # incase of the min_indent_num is 2
        if indent_num > 0 and indent_num < min_indent_num:
            min_indent_num = indent_num
        depth = indent_num // min_indent_num
        if depth + 1 not in bfs_map:
            bfs_map[depth + 1] = []
        tmp = Section(title, source_file, depth + 1, 0)
        bfs_map[depth + 1].append(tmp)
        parent = bfs_map[depth][-1]
        tmp.set_parent(parent)
        tmp.index = len(parent.children)
        parent.add_children(tmp)

    return root




import os
from mdbook_pdf_summary import parse_section_tree, check_title
if __name__ == "__main__":
    summary_path = "./example/a_book/src/SUMMARY.md"
    with open(summary_path) as f:
        md_text = f.read()
    prefix_path = os.path.dirname(summary_path)
    section_tree = parse_section_tree(md_text)
    check_result = check_title(prefix_path, section_tree, overwrite=True)
    print(f"Check result: {check_result}")


from mdbook_pdf_summary.parse_section import Section
import os
def check_title(prefix_path: str, node: Section, overwrite: bool) -> bool:
    """
    Based on `SUMMARY.md`, check whether the title of each markdown file is the same as the content.
    If `overwrite==True`, generate the title at the first line, for those unmatched files.

    Parameters:
    -----------
    prefix_path: str
        The path prefix of the markdown files.
    node: Section
        The node of the section tree.
    overwrite: bool
        Whether to overwrite the title in the markdown file.

    Returns:
    --------
    bool
        Whether all titles are matched (after overwrite).
    """
    all_matched = True
    for child in node.children:
        child_result = check_title(prefix_path, child, overwrite)
        if not child_result:
            return False
    if node.is_root():
        return True
    
    source_file = os.path.join(prefix_path, node.source_file)
    if not os.path.exists(source_file):
        print(f"File {source_file} does not exist")
        return False
    with open(source_file, 'r') as f:
        lines = f.readlines()
    # Find the first title `# ${title}`
    for idx, line in enumerate(lines):
        if line.startswith('# '):
            title = line[2:]
            if not title.startswith(node.title):
                all_matched = False
                print("[ERROR] Title not matched: source_file:{}, line num:{}, title:{}, title in `SUMMARY.md`:{}".format(
                    source_file, idx, title, node.title))
                break
    if not all_matched and overwrite:
        lines.insert(0, f"# {node.title}\n")
        print(f"[Info] Overwrite title as {node.title } in {node.source_file}")
        with open(source_file, 'w') as f:
            f.writelines(lines)
        all_matched = True
    return all_matched




from setuptools import setup, find_packages
import os
this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()
setup(
    name="mdbook_pdf_summary",
    version="0.1.1",
    author="cmq2525",
    author_email="cmq2525@foxmail.com",
    description="Generate outline for pdf generated by `mdbook-pdf`.",
    packages=find_packages(),
    install_requires=["lxml==5.2.1", "pypdf==4.2.0"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={
        "console_scripts": [
            "mdbook-pdf-summary=mdbook_pdf_summary.mdbook_pdf_summary:main"
        ]
    },
    long_description=long_description,
    long_description_content_type="text/markdown",
)
